/* eslint-env node */
'use strict';

process.env.NODE_TLS_REJECT_UNAUTHORIZED = '0';

const fs = require('fs');
const path = require('path');
const url = require('url');
const crypto = require('crypto');
const https = require('https');
const express = require('express');
const cors = require('cors');
const httpProxy = require('http-proxy');

let server = null;
let serverOptions = null;
let useCache = true;

function hashValue(input) {
    const hash = crypto.createHash('sha256').update(input).digest('base64');
    return (new Buffer(hash)).toString('base64');
}
function filterObject(obj, ignoredKeys) {
    const result = {};
    const caseInsensitiveKeys = ignoredKeys.map((key) => key.toLowerCase());
    Object.keys(obj).forEach((key) => {
        if(caseInsensitiveKeys.indexOf(key.toLowerCase()) === -1) {
            result[key] = obj[key];
        }
    });
    return result;
}

process.argv.forEach((arg) => {
    if(arg === '--no-cache') {
        useCache = false;
    }
});

module.exports = {
    run(serverConfig) {
        const expressApp = express();
        const proxy = httpProxy.createProxyServer({ secure: false });
        const defaultConfig = JSON.parse(fs.readFileSync(path.join(__dirname, 'config.json'), 'utf8'));
        const sslOptions = {
            key: fs.readFileSync(path.join(__dirname, 'ssl', 'server.key'), 'utf8'),
            cert: fs.readFileSync(path.join(__dirname, 'ssl', 'server.crt'), 'utf8')
        };
        if(typeof serverConfig !== 'object' || serverConfig === null) {
            serverConfig = {};
        }
        Object.keys(defaultConfig).forEach((key) => {
            if(!(key in serverConfig)) {
                serverConfig[key] = defaultConfig[key];
            }
        });
        serverOptions = serverConfig;
        server = https.createServer(sslOptions, expressApp);
        if('clientOrigin' in serverConfig) {
            expressApp.use(cors({ origin: serverConfig.clientOrigin, credentials: true }));
        }
        expressApp.use((req, res) => {
            const filteredReqHeaders = filterObject(req.headers, serverConfig.ignoredRequestHeaders);
            const reqHeadersString = Object.keys(filteredReqHeaders).sort().map((headerKey) => req.headers[headerKey]).join(':');
            const cacheName = hashValue(`${req.method}:${req.url}:${reqHeadersString}`);
            req.cachePath = path.resolve(serverConfig.cacheDirectory, `${cacheName}.json`);
            fs.readFile(req.cachePath, 'utf8', (err, data) => {
                if(err || useCache === false) {
                    const targetHref = url.parse(req.url, true).query['fetch-recorder-target'];
                    if(typeof targetHref === 'undefined') {
                        res.statusCode = 500;
                        res.end(`fetch-recorder error: could not parse target url\n${req.url}`);
                        return;
                    }
                    const targetUrl = url.parse(targetHref);
                    req.headers.host = targetUrl.host;
                    proxy.web(req, res, { target: targetUrl.href });
                } else {
                    const resData = JSON.parse(data);
                    res.writeHead(resData.statusCode, resData.headers);
                    res.end(resData.body, 'base64');
                }
            });
        });
        proxy.on('proxyRes', (proxyRes, req, res) => {
            const chunks = [];
            proxyRes.on('data', (chunk) => {
                chunks.push(chunk);
            });
            proxyRes.on('end', () => {
                const resData = {
                    statusCode: res.statusCode,
                    headers: filterObject(res._headers, serverConfig.ignoredResponseHeaders),
                    body: Buffer.concat(chunks).toString('base64')
                };
                fs.writeFile(req.cachePath, JSON.stringify(resData), (err) => {
                    if(err) {

                        // eslint-disable-next-line no-console
                        console.log(err);
                    }
                });
            });
        });
        server.listen(serverConfig.port);
    },
    getInfo() {
        if(server === null) {
            throw new Error('fetch-recorder server has not been started');
        }
        return {
            cacheDirectory: path.resolve(serverOptions.cacheDirectory),
            ignoredRequestHeaders: serverOptions.ignoredRequestHeaders,
            ignoredResponseHeaders: serverOptions.ignoredResponseHeaders,
            port: server.address().port,
            targets: serverOptions.targets
        };
    }
};
